from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestAnmeldungMaus:
    """
    A unit test that ensures that the 11001/11002/11003 MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11001.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11001(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11001.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11001_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11002.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11002(self, datafiles):
        result = create_maus_and_assert(
            csv_path=Path(datafiles) / "11002.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11002_maus.json"),
        )
        nav_segments = result.maus.find_segments(
            group_predicate=lambda group: group.discriminator == "SG10",
            segment_predicate=lambda seg: seg.discriminator == "CAV"
            and seg.section_name == "Netznutzungsabrechnungsvariante",
        )
        assert len(nav_segments) == 1  # https://github.com/Hochfrequenz/edifact-templates/issues/73
        for expected_section_name, unexpected_section_name in [
            (
                # https://github.com/Hochfrequenz/edifact-templates/issues/82
                "OBIS-Kennzahl der Zähleinrichtung / Mengenumwerter",
                "OBIS-Kennzahl der Zähleinrichtung / Mengenumwerter / Smartmeter-Gateway",
            ),
            (
                # https://github.com/Hochfrequenz/edifact-templates/issues/82
                "OBIS-Daten der Zähleinrichtung / Mengenumwerter",
                "OBIS-Daten der Zähleinrichtung / Mengenumwerter / Smartmeter-Gateway",
            ),
            (
                # https://github.com/Hochfrequenz/edifact-templates/issues/80
                "OBIS Daten für Marktrolle relevant",
                "OBIS Daten für Lieferant relevant",
            ),
            (
                # https://github.com/Hochfrequenz/edifact-templates/issues/73
                "Netznutzungsabrechnungsvariante",
                "Netznutzungsabrechnungsv ariante",
            ),
        ]:
            expected_zaehleinrichtungs_obis = result.maus.find_segments(
                segment_predicate=lambda seg: seg.section_name == expected_section_name
            )
            assert len(expected_zaehleinrichtungs_obis) > 0
            unexpected_zaehleinrichtungs_obis = result.maus.find_segments(
                segment_predicate=lambda seg: seg.section_name == unexpected_section_name
            )
            assert len(unexpected_zaehleinrichtungs_obis) == 0
        assert (
            len(
                result.maus.find_segments(
                    segment_predicate=lambda seg: seg.discriminator == "SEQ"
                    and seg.section_name == "OBIS-Daten der Marktlokation",
                )
            )
            > 0
        )  # https://github.com/Hochfrequenz/edifact-templates/issues/81
        assert (
            len(
                result.maus.find_segments(
                    segment_predicate=lambda seg: seg.discriminator == "CAV"
                    and seg.section_name == "Druckebene der Marktlokation",
                )
            )
            > 0
        )  # https://github.com/Hochfrequenz/edifact-templates/issues/79

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11003.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11003(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11003.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11003_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11004.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11004(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11004.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11004_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11005.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11005(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11005.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11005_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11006.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11006(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11006.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11006_maus.json"),
        )



"""
Classes that allow to read XML files that contain structural information (Message Implementation Guide information)
"""
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, List, Literal, Optional, TypeVar, Union
from xml.etree.ElementTree import Element

from lxml import etree  # type:ignore[import]

from maus import SegmentGroupHierarchy
from maus.edifact import EdifactFormat
from maus.models._internal import EdifactStackSearchStrategy, MigFilterResult
from maus.models.edifact_components import EdifactStack, EdifactStackLevel, EdifactStackQuery
from maus.reader.etree_element_helpers import (
    filter_by_name,
    filter_by_section_name,
    get_ahb_name_or_none,
    get_nested_qualifiers,
    get_segment_group_key_or_none,
    list_to_mig_filter_result,
)
from maus.reader.mig_ahb_name_helpers import make_name_comparable, make_tree_names_comparable


class MigReader(ABC):
    """
    A MIG reader is a class that reads Message Implementation Guide (MIG) data from a source
    """

    @abstractmethod
    def to_segment_group_hierarchy(self) -> SegmentGroupHierarchy:
        """
        convert the read data into a segment group hierarchy
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method")

    # pylint:disable=too-many-arguments
    @abstractmethod
    def get_edifact_stack(self, query: EdifactStackQuery) -> Optional[EdifactStack]:
        """
        Returns the edifact stack for the given combination of segment group, key, data element and name
        """
        raise NotImplementedError("The inheriting class has to implement this method")


Result = TypeVar("Result")  #: is a type var to indicate an "arbitrary but same" type in a generic function


def check_file_can_be_parsed_as_mig_xml(file_path: Path) -> None:
    """
    Returns nothing iff the given file is parsable as XML and contains no obvious errors.
    This is not a really sophisticated analysis but just a basic minimal sanity check.
    In case of error an exception is raised.
    """
    reader = MigXmlReader(file_path)
    _ = EdifactFormat(reader.get_format_name())  # dies with an exception if the value is invalid


# pylint:disable=c-extension-no-member
class MigXmlReader(MigReader):
    """
    Reads an XML file an transforms it into a Segment Group Hierarchy
    """

    def __init__(self, init_param: Union[str, Path]):
        self._original_root: etree._Element
        if isinstance(init_param, str):
            self._original_root = etree.fromstring(init_param)
            self._sanitized_root = etree.fromstring(init_param)
        elif isinstance(init_param, Path):
            self._original_root = etree.parse(str(init_param.absolute())).getroot()
            self._sanitized_root = etree.parse(str(init_param.absolute())).getroot()
        else:
            raise ValueError(f"The type of '{init_param}' is not valid")
        # the original tree is the unmodified MIG XML Structure with all its quircks
        self._original_tree: etree.ElementTree = etree.ElementTree(self._original_root)
        # but turns out it's much easier to handle a sanitized tree that is simplified in a sense that
        # * has only lower case (ahb)names which are easy to match because they don't contain whitespace,"-" or casing
        # * has the same structure as the _original_tree so that absolute path expressions from the sanitized tree match
        self._sanitized_tree: etree.ElementTree = etree.ElementTree(self._sanitized_root)
        make_tree_names_comparable(self._sanitized_tree)

    def get_format_name(self) -> str:
        """
        the root element of the XML is the name of the EDIFACT format
        """
        return self._original_root.tag

    def element_to_edifact_stack(self, element: etree.Element, use_sanitized_tree: bool) -> EdifactStack:
        """
        extract the edifact seed path from the given element.
        """
        # this method is directly unittests. Please refer to the test for some easy to debug examples.
        if use_sanitized_tree:
            xpath = self._sanitized_tree.getpath(element)
        else:
            xpath = self._original_tree.getpath(element)
        stack = EdifactStack(levels=[])
        iter_path = "/" + xpath.split("/")[1]
        for leaf in xpath.split("/")[2:]:
            iter_path += "/" + leaf
            # here we _always_ need to use the original root!
            leaf_element = self._original_root.xpath(iter_path)[0]  # type:ignore[attr-defined]
            level_name: str
            if "ahbName" in leaf_element.attrib:
                level_name = leaf_element.attrib["ahbName"]
            else:
                level_name = leaf_element.attrib["name"]
            # https://stackoverflow.com/questions/47972143/using-attr-with-pylint
            # pylint: disable=no-member
            stack.levels.append(EdifactStackLevel(name=level_name, is_groupable=leaf_element.tag == "class"))
        return stack

    @staticmethod
    def get_unique_result_by_name(candidates: List[Element], query: EdifactStackQuery) -> MigFilterResult:
        """
        returns those elements that have the given name
        """
        # this method is directly unittests. Please refer to the test for some easy to debug examples.
        filtered_by_names = filter_by_name(candidates, query)
        return list_to_mig_filter_result(filtered_by_names)

    @staticmethod
    def get_unique_result_by_section_name(candidates: List[Element], query: EdifactStackQuery) -> MigFilterResult:
        """
        keeps those elements from the candidates whose where the name matches the query section name.
        Does _not_ create a new xpath.
        """
        # this method is directly unittests. Please refer to the test for some easy to debug examples.
        filtered_by_names = filter_by_section_name(candidates, query)
        return list_to_mig_filter_result(filtered_by_names)

    def get_unique_result_by_xpath(self, query_path: str, use_sanitized_tree: bool) -> MigFilterResult:
        """
        Tries to find an element for the given query path.
        If there's exactly 1 result, it is returned.
        If there are 0 or >1 results a ValueError is raised.
        """
        candidates: List[Element]
        if use_sanitized_tree:
            candidates = list(self._sanitized_root.xpath(query_path))
        else:
            candidates = list(self._original_root.xpath(query_path))
        return list_to_mig_filter_result(candidates)

    def get_unique_result_by_segment_group(
        self, candidates: List[Element], query: EdifactStackQuery, use_sanitized_tree: bool
    ) -> MigFilterResult:
        """
        keep those elements that have the correct segment_group_key
        """
        # this method is separately unittests; see the tests to get an understanding of the way it works.
        filtered = [
            e
            for e in candidates
            if self.get_parent_segment_group_key(e, use_sanitized_tree=use_sanitized_tree) == query.segment_group_key
        ]
        return list_to_mig_filter_result(filtered)

    # pylint:disable=no-self-use
    def get_unique_result_by_predecessor(self, candidates: List[Element], query: EdifactStackQuery) -> MigFilterResult:
        """
        Keep those elements that have (in the field) the given predecessor qualifier
        """
        # this method is separately unittests; see the tests to get an understanding of the way it works.
        relevant_attribute: Literal["key", "ref"]
        if query.segment_code == "RFF":
            relevant_attribute = "key"
        else:
            relevant_attribute = "ref"
        result = []
        for candidate in candidates:
            nested_qualifiers = get_nested_qualifiers(relevant_attribute, candidate)
            if nested_qualifiers is not None and query.predecessor_qualifier in nested_qualifiers:
                # that's a bit dirty, better parse the ref properly instead of string-matching
                result.append(candidate)
        return list_to_mig_filter_result(result)

    def get_unique_result_by_parent_predecessor(
        self, candidates: List[Element], query: EdifactStackQuery, skip_levels: int = 0
    ) -> MigFilterResult:
        """
        Keep those elements that have (in the parent class) the given predecessor qualifier
        """
        # this method is separately unittests; see the tests to get an understanding of the way it works.
        result = []
        for candidate in candidates:
            parent_predecessors = self.get_parent_predecessors(
                candidate, use_sanitized_tree=False, skip_levels=skip_levels
            )
            if parent_predecessors is not None and query.predecessor_qualifier in parent_predecessors:
                result.append(candidate)
        return list_to_mig_filter_result(result)

    def get_unique_result_by_parent_segment_group(
        self, candidates: List[Element], query: EdifactStackQuery
    ) -> MigFilterResult:
        """
        Keep those elements that have (in the parent class) the given segment group key
        """
        filtered_by_segment_group_key = [
            c
            for c in candidates
            if self.get_parent_segment_group_key(c, use_sanitized_tree=False) == query.segment_group_key
        ]
        return list_to_mig_filter_result(filtered_by_segment_group_key)

    def get_unique_result_by_parent_ahb_name_section_name(
        self, candidates: List[Element], query: EdifactStackQuery
    ) -> MigFilterResult:
        """
        Keep those elements that have (in the parent class) the given query.section_name
        """
        filtered_by_parent_ahb_name_key = [
            c
            for c in candidates
            if make_name_comparable(self._get_parent_x(c, get_ahb_name_or_none, use_sanitized_tree=False))
            == make_name_comparable(query.section_name)
        ]
        return list_to_mig_filter_result(filtered_by_parent_ahb_name_key)

    def _get_parent_x(
        self, element: Element, evaluator: Callable[[Element], Result], use_sanitized_tree: bool, skip_levels: int = 0
    ) -> Optional[Result]:
        """
        get the 'X' property of the parent where 'X' is the result of the evaluator when applied to an element.
        returns None if not found
        """
        result = evaluator(element)
        if result is not None:
            return result
        if use_sanitized_tree:
            xpath = self._sanitized_tree.getpath(element)
        else:
            xpath = self._original_tree.getpath(element)
        path_parts = list(xpath.split("/"))
        sub_paths: List[str] = []
        for depth in range(2, len(path_parts) - skip_levels):
            sub_path = "/".join(path_parts[0:depth])
            sub_paths.append(sub_path)
        # if xpath was "/foo/bar/asd/xyz", sub_paths is ["/foo", "/foo/bar", "/foo/bar/asd"] now (xyz is not contained!)
        sub_paths.reverse()
        for sub_path in sub_paths:  # loop from inner to root
            leaf_element = self._original_root.xpath(sub_path)[0]  # type:ignore[attr-defined]
            result = evaluator(leaf_element)
            if result is not None:
                return result
        return None

    def get_parent_segment_group_key(self, element: Element, use_sanitized_tree: bool) -> Optional[str]:
        """
        iterate from element towards root and return the first segment group found (the one closes to element).
        returns None if no segment group was found
        """
        # This method is separately unit tested.
        # Reading the test will most likely make its behaviour more understandable.
        return self._get_parent_x(element, get_segment_group_key_or_none, use_sanitized_tree=use_sanitized_tree)

    def get_parent_predecessors(
        self, element: Element, use_sanitized_tree: bool, skip_levels: int = 0
    ) -> Optional[List[str]]:
        """
        iterate from element towards root and return the first segment group found (the one closes to element).
        returns None if no segment group was found
        """
        # This method is separately unit tested.
        # Reading the test will most likely make its behaviour more understandable.
        result = self._get_parent_x(
            element,
            lambda c: get_nested_qualifiers("key", c),
            use_sanitized_tree=use_sanitized_tree,
            skip_levels=skip_levels,
        )
        return result

    def _handle_predecessor_if_present(self, query: EdifactStackQuery) -> Optional[EdifactStackSearchStrategy]:
        """
        return a strategy for the predecessor if it's present in the query
        """
        if not query.predecessor_qualifier:
            return None
        return EdifactStackSearchStrategy(
            name="I filter by predecessor",
            filter=lambda q, c: self.get_unique_result_by_predecessor(
                c, q  # type:ignore[arg-type]
            ),
            unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                unique_result, use_sanitized_tree=False
            ),
            no_result_strategy=EdifactStackSearchStrategy(
                name="J filter by parent predecessor after predecessor lead to no result",
                filter=lambda q, c: self.get_unique_result_by_parent_predecessor(
                    c, q  # type:ignore[arg-type]
                ),
                no_result_strategy=EdifactStackSearchStrategy(
                    name="M filter by section_name==field name",
                    filter=lambda q, c: MigXmlReader.get_unique_result_by_section_name(c, q),  # type:ignore[arg-type]
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=False
                    ),
                    no_result_strategy=None,
                    multiple_results_strategy=None,
                ),
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                multiple_results_strategy=EdifactStackSearchStrategy(
                    name="K filter by parents segment group",
                    filter=lambda q, c: self.get_unique_result_by_parent_segment_group(
                        c, q  # type:ignore[arg-type]
                    ),
                    no_result_strategy=None,  # filter by its own predecessor :-(
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=False
                    ),
                    multiple_results_strategy=None,
                ),  # Z18 goes here
            ),
            multiple_results_strategy=EdifactStackSearchStrategy(
                # doe snot work yet
                name="L filter by parents segment group because parent predecessor is not unique",
                filter=lambda q, c: self.get_unique_result_by_parent_segment_group(c, q),  # type:ignore[arg-type]
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                multiple_results_strategy=None,
                no_result_strategy=None,
            ),
        )

    def _multiple_data_element_matches_handling(self, query: EdifactStackQuery) -> Optional[EdifactStackSearchStrategy]:
        if query.name is not None:
            return EdifactStackSearchStrategy(
                name="A filter by element name and ahb name",
                filter=lambda q, c: self.get_unique_result_by_name(c, q),  # type:ignore[arg-type]
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                no_result_strategy=EdifactStackSearchStrategy(
                    name="B filter by parents name because direct name lead to no result",
                    filter=lambda q, _: self.get_unique_result_by_xpath(
                        # pylint:disable=line-too-long
                        f".//class[@name='{make_name_comparable(query.name)}']/*[@meta.id='{query.data_element_id}' and starts-with(@ref, '{query.segment_code}')]",  # type:ignore[arg-type]
                        use_sanitized_tree=True,
                    ),
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=True
                    ),
                    no_result_strategy=EdifactStackSearchStrategy(
                        name="C filter by parents ahb name",
                        filter=lambda q, _: self.get_unique_result_by_xpath(
                            # pylint:disable=line-too-long
                            f".//class[@ahbName='{make_name_comparable(query.name)}']/*[@meta.id='{query.data_element_id}' and starts-with(@ref, '{query.segment_code}')]",  # type:ignore[arg-type]
                            use_sanitized_tree=True,
                        ),
                        unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                            unique_result, use_sanitized_tree=True
                        ),
                        multiple_results_strategy=None,
                        no_result_strategy=self._handle_predecessor_if_present(query),
                    ),
                    multiple_results_strategy=None,
                ),
                multiple_results_strategy=EdifactStackSearchStrategy(
                    name="D filter by segment group",
                    filter=lambda q, c: self.get_unique_result_by_segment_group(
                        c, q, use_sanitized_tree=False  # type:ignore[arg-type]
                    ),
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=False
                    ),
                    multiple_results_strategy=EdifactStackSearchStrategy(
                        name="E filter by predecessor after segment group filter was not unique",
                        filter=lambda q, c: self.get_unique_result_by_predecessor(
                            c, q  # type:ignore[arg-type]
                        ),
                        unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                            unique_result, use_sanitized_tree=False
                        ),
                        multiple_results_strategy=None,  # raise ValueError("Predecessor B"),
                        no_result_strategy=EdifactStackSearchStrategy(
                            name="F filter by parent predecessor after predecessor lead to no result",
                            filter=lambda q, c: self.get_unique_result_by_parent_predecessor(
                                c, q  # type:ignore[arg-type]
                            ),
                            no_result_strategy=None,
                            multiple_results_strategy=None,
                            unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                                unique_result, use_sanitized_tree=False
                            ),
                        ),
                    ),
                    no_result_strategy=None,
                ),
            )
        if query.name is None and query.predecessor_qualifier is not None:
            return EdifactStackSearchStrategy(
                name="G filter by parent predecessor",
                filter=lambda q, c: self.get_unique_result_by_parent_predecessor(
                    c, q  # type:ignore[arg-type]
                ),
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                no_result_strategy=EdifactStackSearchStrategy(
                    name="H filter by predecessor after parent predecessor lead to no result",
                    filter=lambda q, c: self.get_unique_result_by_predecessor(
                        c, q  # type:ignore[arg-type]
                    ),
                    unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                        unique_result, use_sanitized_tree=False
                    ),
                    multiple_results_strategy=None,
                    no_result_strategy=EdifactStackSearchStrategy(
                        name="O filter by predecessor after parent predecessor with and skip inner level",
                        filter=lambda q, c: self.get_unique_result_by_parent_predecessor(
                            c, q, skip_levels=1  # type:ignore[arg-type]
                        ),
                        unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                            unique_result, use_sanitized_tree=False
                        ),
                        multiple_results_strategy=None,
                        no_result_strategy=None,
                    ),
                ),
                multiple_results_strategy=None,
            )
        if query.section_name is not None:
            return EdifactStackSearchStrategy(
                name="N filter by parents ahb name (using the section name)",
                filter=lambda q, c: self.get_unique_result_by_parent_ahb_name_section_name(
                    c, q  # type:ignore[arg-type]
                ),
                unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                    unique_result, use_sanitized_tree=False
                ),
                multiple_results_strategy=None,
                no_result_strategy=None,
            )
        return None

    def get_edifact_stack(self, query: EdifactStackQuery) -> Optional[EdifactStack]:
        """
        get the edifact stack for the given segment_group, segment... combination or None if there is no match
        """
        strategy = EdifactStackSearchStrategy(
            name="filter by data element id",
            filter=lambda q, _: (
                self.get_unique_result_by_xpath(
                    f".//*[@meta.id='{q.data_element_id}' and starts-with(@ref, '{q.segment_code}')]",
                    use_sanitized_tree=False,
                )
            ),
            unique_result_strategy=lambda unique_result: self.element_to_edifact_stack(
                unique_result, use_sanitized_tree=False
            ),
            no_result_strategy=None,  # if there is no matching data element, there's no chance to find a stack
            multiple_results_strategy=self._multiple_data_element_matches_handling(query),
        )
        return strategy.apply(query)

    def to_segment_group_hierarchy(self) -> SegmentGroupHierarchy:
        """
        convert the read data into a segment group hierarchy
        :return:
        """
        raise NotImplementedError("The inheriting class has to implement this method")



from typing import Optional

import pytest  # type:ignore[import]
from lxml import etree  # type:ignore[import]

from maus.reader.mig_ahb_name_helpers import are_similar_names, make_name_comparable, make_tree_names_comparable


class TestMigAhbNameHelpers:
    """
    Tests the behaviour of static helper methods
    """

    @pytest.mark.parametrize(
        "x,expected_result",
        [
            pytest.param("", ""),
            pytest.param(" foo", "foo", id="leading space"),
            pytest.param("foo ", "foo", id="trailing space"),
            pytest.param("Foo ", "foo", id="title case"),
            pytest.param("fOo ", "foo", id="mocking spongebob case"),
            pytest.param("🐈Miau ", "🐈miau", id="emoji"),
            pytest.param("Foo-Bar ", "foobar", id="minus"),
            pytest.param("Foo Bar ", "foobar", id="words"),
            pytest.param("Foo\nBar ", "foobar", id="new line"),
        ],
    )
    def test_make_name_comparable(self, x: str, expected_result: str):
        actual = make_name_comparable(x)
        assert actual == expected_result

    @pytest.mark.parametrize(
        "x,y, expected_result",
        [
            pytest.param(None, "", True),
            pytest.param("", "", True),
            pytest.param("Foo", "", False),
            pytest.param(None, "Bar", False),
            pytest.param("X", "X", True),
            pytest.param("X", "x", True),
            pytest.param("X", "y", False),
            pytest.param("Gültigkeit, Beginndatum", "Gültigkeit,Beginndatum", True),
            pytest.param(
                "Referenz Vorgangsnummer (aus Anfragenachricht)", "Referenz Vorgangsnummer (aus Anfragenachricht)", True
            ),
        ],
    )
    def test_are_similar_names(self, x: Optional[str], y: Optional[str], expected_result: bool):
        actual = are_similar_names(x, y)
        assert actual == expected_result

    def test_make_tree_names_comparable(self):
        orig_xml = """<?xml version="1.0"?>
        <hello>
        <einfield name="Foo "/>
        <einclass ahbName="Hallo-Welt">
            <einfield name=" DiGiTaLiSiErUnG" ahbName="(ZuKunFt)"/>
        </einclass>
        </hello>
        """
        tree = etree.ElementTree(etree.fromstring(orig_xml))
        make_tree_names_comparable(tree)
        for element in tree.iter():
            if "name" in element.attrib:
                assert element.attrib["name"] == make_name_comparable(element.attrib["name"])
            if "ahbName" in element.attrib:
                assert element.attrib["ahbName"] == make_name_comparable(element.attrib["ahbName"])



from pathlib import Path

import pytest  # type:ignore[import]
from helpers import create_maus_and_assert, write_to_file_or_assert_equality  # type:ignore[import]


class TestStammdatenaenderungMaus:
    """
    A unit test that ensures that the 111xx MAUS.json are created.
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11116.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11116(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11116.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11116_maus.json"),
        )

    """
    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11117.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11117(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11117.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11117_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11119.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11119(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11119.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11119_maus.json"),
        )
    """

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11168.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11168(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11168.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11168_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11169.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11169(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11169.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11169_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11170.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11170(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11170.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11170_maus.json"),
        )

    @pytest.mark.datafiles("./edifact-templates/edi/UTILMD/UTILMD5.2c.template")
    @pytest.mark.datafiles("./edifact-templates/ahbs/FV2110/UTILMD/11194.csv")
    @pytest.mark.datafiles("../unit_tests/migs/FV2204/segment_group_hierarchies/sgh_utilmd.json")
    def test_maus_creation_11194(self, datafiles):
        create_maus_and_assert(
            csv_path=Path(datafiles) / "11194.csv",
            sgh_path=Path(datafiles) / "sgh_utilmd.json",
            template_path=Path(datafiles) / Path("UTILMD5.2c.template"),
            maus_path=Path("edifact-templates/maus/FV2110/UTILMD/11194_maus.json"),
        )



from typing import Callable

import pytest  # type:ignore[import]

from maus import DeepAnwendungshandbuch, Segment, SegmentGroup
from maus.models.anwendungshandbuch import AhbMetaInformation


class TestSearchingInModels:
    """
    Tests regarding the search feature in Edifact Components or AHBs.
    """

    @pytest.mark.parametrize(
        "segment_group,search_predicate,number_of_results",
        [
            pytest.param(
                SegmentGroup(discriminator="SG1", ahb_expression="Foo"),
                lambda _: False,
                0,
                id="always False predicate leads to no results",
            ),
            pytest.param(
                SegmentGroup(discriminator="SG1", ahb_expression="Foo"), lambda _: True, 0, id="no segments, no results"
            ),
            pytest.param(
                SegmentGroup(
                    discriminator="SG1",
                    ahb_expression="Foo",
                    segments=[Segment(discriminator="ASD", data_elements=[], ahb_expression="Foo")],
                ),
                lambda _: True,
                1,
                id="true predicate",
            ),
            pytest.param(
                SegmentGroup(
                    discriminator="SG1",
                    ahb_expression="Foo",
                    segments=[Segment(discriminator="ASD", data_elements=[], ahb_expression="Foo")],
                ),
                lambda seg: seg.discriminator == "ASD",
                1,
                id="matching predicate",
            ),
            pytest.param(
                SegmentGroup(
                    discriminator="SG1",
                    ahb_expression="Foo",
                    segments=[Segment(discriminator="Foo", data_elements=[], ahb_expression="Foo")],
                ),
                lambda seg: seg.discriminator == "Bar",
                0,
                id="not matching predicate",
            ),
            pytest.param(
                SegmentGroup(
                    discriminator="SG1",
                    ahb_expression="Foo",
                    segments=[
                        Segment(discriminator="Foo", data_elements=[], ahb_expression="Foo"),
                        Segment(discriminator="Bar", data_elements=[], ahb_expression="Bar"),
                    ],
                ),
                lambda seg: seg.discriminator == "Bar",
                1,
                id="matching predicate on root level",
            ),
            pytest.param(
                SegmentGroup(
                    discriminator="SG1",
                    ahb_expression="Foo",
                    segments=[
                        Segment(discriminator="abc", ahb_expression="ahb expression", data_elements=[]),
                        Segment(
                            discriminator="abc",
                            ahb_expression="something not ending with expression but foo",
                            data_elements=[],
                        ),
                    ],
                    segment_groups=[
                        SegmentGroup(
                            discriminator="SG2",
                            ahb_expression="Bar",
                            segments=[
                                Segment(discriminator="abc", ahb_expression="ahb expression", data_elements=[]),
                                Segment(discriminator="xyz", ahb_expression="xyz expression", data_elements=[]),
                            ],
                        )
                    ],
                ),
                lambda seg: seg.ahb_expression.endswith("expression"),
                3,
                id="matching predicate with 3 entries",
            ),
        ],
    )
    def test_find_segment_in_segment_group(
        self, segment_group: SegmentGroup, search_predicate: Callable[[Segment], bool], number_of_results: int
    ):
        actual = segment_group.find_segments(search_predicate)
        assert len(actual) == number_of_results

    @pytest.mark.parametrize(
        "deep_ahb,search_predicate,number_of_results",
        [
            pytest.param(
                DeepAnwendungshandbuch(
                    lines=[
                        SegmentGroup(discriminator="SG1", ahb_expression="Foo"),
                        SegmentGroup(
                            discriminator="SG2",
                            ahb_expression="Bar",
                            segment_groups=[SegmentGroup(discriminator="SG3", ahb_expression="Baz")],
                        ),
                    ],
                    meta=AhbMetaInformation(pruefidentifikator="11111"),
                ),
                lambda sg: sg.discriminator == "SG3",
                1,
            ),
        ],
    )
    def test_find_segment_group_from_deep_ahb(
        self, deep_ahb: DeepAnwendungshandbuch, search_predicate: Callable[[SegmentGroup], bool], number_of_results: int
    ):
        actual = deep_ahb.find_segment_groups(search_predicate)
        assert len(actual) == number_of_results

    @pytest.mark.parametrize(
        "deep_ahb,group_predicate,segment_predicate,number_of_results",
        [
            pytest.param(
                DeepAnwendungshandbuch(
                    lines=[
                        SegmentGroup(discriminator="SG1", ahb_expression="Foo"),
                        SegmentGroup(
                            discriminator="SG2",
                            ahb_expression="Bar",
                            segment_groups=[
                                SegmentGroup(
                                    discriminator="SG3",
                                    ahb_expression="Baz",
                                    segments=[
                                        Segment(discriminator="abc", ahb_expression="ahb expression", data_elements=[]),
                                        Segment(
                                            discriminator="qwertz", ahb_expression="qwertz expression", data_elements=[]
                                        ),
                                    ],
                                )
                            ],
                        ),
                    ],
                    meta=AhbMetaInformation(pruefidentifikator="11111"),
                ),
                lambda sg: sg.discriminator == "SG3",
                lambda seg: len(seg.discriminator) > 3,
                1,
            ),
        ],
    )
    def test_find_segments_from_deep_ahb(
        self,
        deep_ahb: DeepAnwendungshandbuch,
        group_predicate: Callable[[SegmentGroup], bool],
        segment_predicate: Callable[[Segment], bool],
        number_of_results: int,
    ):
        actual = deep_ahb.find_segments(group_predicate, segment_predicate)
        assert len(actual) == number_of_results
